var Protocol = function()
{
	var protocol = {};
	var debug = 0;
	
	var pools;
	var libPlate;

	var POOL_CURRENT_SCHEMA;
	var POOL_CURRENT_BARCODE_VARIANT;

	// Page initialization
	protocol.initPage = function()
	{
		var pageId = Doc.getPageId();
		if (pageId == 'protocol')
		{
			Buttons.addClickHandler('print-button', Wizard.goPrint);	
			protocol.initializeProtocol();
		}
		else
		{
			Events.addEventHandler('bioplate', 'change', protocol.bioplateOnChange);
			Events.addEventHandler('listview', 'click', protocol.viewProtocol);
			Events.addEventHandler('plateview', 'click', protocol.viewProtocol);

			var url = '../Pool.servlet?ID='+App.getSessionId();
			url += '&cmd=GetUnprocessedPools';
			Wizard.showLoadingAnimation('Loading pools...');
			Wizard.asyncJsonRequest(url, protocol.initializeStep1);
		}
	}


	protocol.initializeStep1 = function(response)
	{
		var frm = document.forms['reggie'];
		pools = response.pools;
		
		// We get a list of pools from the servlet
		// Each pool has a list of libraries 
		// We want to find the Library bioplates that are unique
		// within a pool (eg. all libraries are from the same bioplate)
		if (pools.length == 0)
		{
			Wizard.setFatalError('No pools available for processing.');
			return;
		}
		
		var poolsList = frm.pools;
		var plateList = frm.bioplate;
		var libPlates = [];
		for (var poolNo=0; poolNo < pools.length; poolNo++)
		{
			var pool = pools[poolNo];
			
			// Shortcut to single-plate pools
			var libPlateForPool = pool.libPlates.length == 1 && !pool.isManual ? pool.libPlates[0] : null;
			pool.libPlate = libPlateForPool;

			// If the pool has a single library bioplate add it to the selection list
			// If the pool doesn't have a single library bioplate, nothing is added to
			// the list and the pool can only be selected when the 'manual pools' option is
			// active for the bioplates list
			if (libPlateForPool != null && !libPlates[libPlateForPool.name])
			{
				libPlates[libPlateForPool.name] = libPlateForPool;
				var plateOption = new Option(libPlateForPool.name, libPlateForPool.id);
				plateOption.plate = libPlateForPool;
				plateList[plateList.length] = plateOption;
			}

		}
		if (plateList.length > 1) plateList.selectedIndex = 1;

		protocol.bioplateOnChange();
		Doc.show('step-1');
		Doc.show('gonext');
	}
	
	
	// Add pools to the pools list based on the bioplate selection
	protocol.bioplateOnChange = function()
	{
		var frm = document.forms['reggie'];
		
		var libPlateId = parseInt(frm.bioplate.value);

		// Clear current list
		frm.pools.length = 0;
		
		for (var poolNo = 0; poolNo < pools.length; poolNo++)
		{
			var pool = pools[poolNo];
			// If a libplate is selected, only show pools that have all libraries from that plate
			// If no libplate is selected, only show pools that have libraries from multiple plates
			var addPool = (libPlateId &&  pool.libPlate && pool.libPlate.id == libPlateId) || (!libPlateId && pool.libPlate == null);
			if (addPool)
			{
				var name = pool.name;
				if (pool.libPlate == null)
				{
					
					name += ' - ';
					var numPlates = pool.libPlates.length;
					var title = '';
					if (numPlates <= 2)
					{
						for (var plateNo=0; plateNo < numPlates; plateNo++)
						{
							var libPlate = pool.libPlates[plateNo];
							if (plateNo > 0) name += ', ';
							name += libPlate.name;
						}
					}
					else
					{
						name += pool.libPlates[0].name + ' + ' + (numPlates-1) + ' more...';
						for (var plateNo=0; plateNo < numPlates; plateNo++)
						{
							var libPlate = pool.libPlates[plateNo];
							if (plateNo > 0) title += ', ';
							title += libPlate.name;
						}
					}
				}

				var option = new Option(name, pool.id, libPlateId || frm.pools.length == 0);
				option.pool = pool;
				frm.pools[frm.pools.length] = option;
			}
		}
		
		Wizard.setInputStatus('pools');
		if (libPlateId)
		{
			frm.pools.disabled = true;
			Doc.removeClass('plateview', 'disabled');
		}
		else
		{
			frm.pools.disabled = false;
			Doc.addClass('plateview', 'disabled');
		}
	}
	
	protocol.viewProtocol = function(event)
	{
		if (event.currentTarget.className.indexOf('disabled')>=0) return;
		
		var frm = document.forms['reggie'];
		var libPlateId = parseInt(frm.bioplate.value);
		
		if (libPlateId == 0)
		{
			// 'manual' option selected, check that at least one pool is selected
			var numSelectedPools = 0;
			for (var poolNo = 0; poolNo < frm.pools.length; poolNo++)
			{
				if (frm.pools[poolNo].selected) numSelectedPools++;
			}
			
			if (numSelectedPools == 0)
			{
				Wizard.setInputStatus('pools', 'invalid', 'Select at least one pool');
				Wizard.notifyError();
				return;
			}
			Wizard.setInputStatus('pools', 'valid');
		}

		var type = Data.get(event.currentTarget, 'protocol-type');
		frm.pools.disabled = false;
		frm.view.value = type;
		frm.submit();
		frm.pools.disabled = (libPlateId != 0);
	}

	protocol.loadLibPlate = function()
	{
		var libPlateId = Data.get('page-data', 'libplate');
		var url = '../Pool.servlet?ID='+App.getSessionId();
		url += '&cmd=GetLibPlateInfo&bioplate='+libPlateId;		
		Wizard.showLoadingAnimation('Loading library bioplate information...');
		Wizard.asyncJsonRequest(url, protocol.initializeProtocol);
	}
	
	protocol.loadPools = function()
	{
		var pools = Data.get('page-data', 'pools');
		var libPlateId = Data.get('page-data', 'libplate');

		var url = '../Pool.servlet?ID='+App.getSessionId();
		url += '&cmd=GetLibraryInfoForPools&pools='+pools+'&bioplate='+libPlateId;

		Wizard.showLoadingAnimation('Loading pooled library information...');
		Wizard.asyncJsonRequest(url, protocol.initializeProtocol);
	}
	
	protocol.initializeProtocol = function(response)
	{
		var libPlateId = Data.int('page-data', 'libplate');
		var view = Data.get('page-data', 'view');

		if (!response)
		{
			if (libPlateId)
			{
				protocol.loadLibPlate(libPlateId);
			}
			else
			{
				protocol.loadPools();
			}
			return;
		}
		else if (response.libPlate)
		{
			libPlate = response.libPlate;
			POOL_CURRENT_SCHEMA = PoolSchema.getById(libPlate.poolSchema);
			POOL_CURRENT_BARCODE_VARIANT = PoolSchema.getBarcodeVariantByName(POOL_CURRENT_SCHEMA, libPlate.barcodeVariant);
			protocol.loadPools();
			return;
		}
		
		pools = response.pools;
		var poolInfo = response.poolInfo;

		var pageTitle = 'Lab protocol for pooling ';
		// Pre-process the Library items
		for (var i = 0; i < pools.length; i++)
		{
			var pool = pools[i];
			if (i > 0) pageTitle += ', '
			pageTitle += ' ' + pool.name;
			
			// Calculate some pool quantities
			pool.volume = 1000 * pool.originalQuantity / pool.conc;

			for (var libNo = 0; libNo < pool.libraries.length; libNo++)
			{
				var lib = pool.libraries[libNo];
				PoolMix.calculateLibVolumeForProtocol(lib, pool.targetPoolMolarity, pool.targetVolumePerLib, pool.mixingStrategy, true);
				PoolMix.calculateEbVolume(lib, pool.targetPoolMolarity, pool.targetVolumePerLib, pool.mixingStrategy);
				protocol.checkAndPreProcessLibrary(lib, pool, POOL_CURRENT_SCHEMA, POOL_CURRENT_BARCODE_VARIANT, view);				
			}

			pool.extra = PoolMix.calculateFinalPoolInfo(pool.libraries, pool.targetPoolMolarity, pool.targetVolumePerLib, pool.mixingStrategy);
		}
		
		if (!libPlate)
		{
			document.title = pageTitle;
		}
		
		if (view == 'list')
		{
			protocol.viewAsList(pools, POOL_CURRENT_SCHEMA, POOL_CURRENT_BARCODE_VARIANT);
		}
		else
		{
			protocol.viewAsPlate(pools, POOL_CURRENT_SCHEMA, POOL_CURRENT_BARCODE_VARIANT);
		}
		
	}
	
	
	protocol.checkAndPreProcessLibrary = function(lib, pool, schema, barcodeVariant, view)
	{
		var well = lib.bioWell;
		var remarks = [];

		lib.isYellow = lib.specimen && lib.specimen.YellowLabel != null;
		
		// Check if default barcode has been modified
		if (barcodeVariant)
		{
			var indexSet = barcodeVariant.indexSets[well.column];
			if (indexSet)
			{
				var defaultBarcode = indexSet.barcodes[well.row];
				if (defaultBarcode && lib.barcode.name != defaultBarcode)
				{
					remarks[remarks.length] = 'Modified barcode';
					lib.barcode.modified = true;
				}
			}
		}
		
		if (lib.molarity)
		{
			if (lib.speedVacConc != null)
			{
				remarks[remarks.length] = 'SpeedVac';
			}
	
			if (lib.mixFactor > 1)
			{
				// Larger mix than default
				if (view != 'plate')
				{
					remarks[remarks.length] = '<span class="mix-remark">Mix <span class="volume">'+Numbers.formatNumber(lib.actualVolume*lib.mixFactor, 1) + '</span>+<span class="eb">' + Numbers.formatNumber(lib.actualEb*lib.mixFactor, 1)+'µl</span></span>';
				}
				remarks[remarks.length] = 'Use <b>' + Numbers.formatNumber(lib.actualVolume+lib.actualEb, 1) + 'µl</b> in pool';
			}

			lib.remarks = remarks;
		}
		
		if (lib.comment) remarks[remarks.length] = lib.comment;
	}

	protocol.viewAsPlate = function(pools, schema, barcodeVariant)
	{
		var isNeoPrep = libPlate.isNeoPrep;
		var columns = libPlate.columns;
		var rows = libPlate.rows;
		
		var html = '<tr id="pool-row"></tr>';
		html += '<tr class="header">';
		html += '<th></th>';
		for (var c = 0; c < columns; c++)
		{
			html += '<th id="col.'+c+'">'+(c+1)+'</th>';
		}
		if (isNeoPrep)
		{
			html += '<th></th>';
		}
		html += '</tr>';
		html += '<tbody>';
		
		for (var r = 0; r < rows; r++)
		{
			html += '<tr class="row-'+r+'">';
			html += '<th id="row.'+r+'" class="rowheader">'+(isNeoPrep ? (r+1)+'L' : Reggie.wellToAlpha(r))+'</th>';

			for (var c = 0; c < columns; c++)
			{
				html += '<td class="well col-'+c+'" id="well.'+r+'.'+c+'"';
				html += ' data-col="'+c+'" data-row="'+r+'"';
				html += ' title="Select/deselect this well"></td>';
			}
			if (isNeoPrep)
			{
				html += '<th id="row.'+r+'.2" class="rowheader">'+(r+9)+'L</th>';
			}
			html += '</tr>';
		}
		html += '</tbody>';
		Doc.element('plateview').innerHTML = html;
		
		Plate.init(rows, columns, schema, WellPainter);

		for (var poolNo = 0; poolNo < pools.length; poolNo++)
		{
			var pool = pools[poolNo];
			for (var libNo = 0; libNo < pool.libraries.length; libNo++)
			{
				var lib = pool.libraries[libNo];
				var well = lib.bioWell;
				Plate.getWell(well.row, well.column).setExtract(lib);
			}
		}

		WellPainter.barcodeVariant = barcodeVariant;
		PoolSchema.buildPoolTableRow(schema, libPlate.columns, false, null, isNeoPrep ? '<th></th>' : null);
		
		Doc.show('all-protocol');
		
		for (var poolNo = 0; poolNo < pools.length; poolNo++)
		{
			var pool = pools[poolNo];
			var realPoolNo = schema.getPoolNumForColumn(pool.libraries[0].bioWell.column);
			
			var poolData = '<div class="pool-data">';
			poolData += pool.libraries.length + ' libs • ';
			poolData += Numbers.formatNumber(pool.molarity, 2)+'nM • ';
			poolData += Numbers.formatNumber(pool.volume, 1) + 'µl';
			if (pool.mixingStrategy == 'dynamic')
			{
				poolData += ' • <span class="pool-eb">'+Numbers.formatNumber(pool.extra.ebVolumeExtra, 1, 'µl')+'</span>';
			}
			poolData += '<br>Mixing strategy: ' + pool.mixingStrategy;
			poolData += '<div class="comments">'+pool.comments+'</div>';
			poolData += '</div>';
			
			Doc.element('pool.'+realPoolNo).innerHTML = Strings.encodeTags(pool.name) + poolData;
		}
		Plate.paint(Plate.getWells());
	}

	protocol.viewAsList = function(pools, schema, barcodeVariant)
	{
		if (libPlate)
		{
			Plate.init(libPlate.rows, libPlate.columns, schema, null);
		}
		var homeUrl = Data.get('page-data', 'home-url');
		var yellowImg = '<img src="'+homeUrl+'/images/yellow-label-small.png">';
		for (var poolNo = 0; poolNo < pools.length; poolNo++)
		{
			var pool = pools[poolNo];
			var wellsInPool = null;
			if (schema)
			{
				var realPoolNo = schema.getPoolNumForColumn(pool.libraries[0].bioWell.column);
				wellsInPool = Plate.getPool(realPoolNo);
			}
			
			var html = '';
			var tbody = Doc.element('listview.'+pool.id+'.body');
			
			var wellIndex = 0;
			for (var j = 0; j < pool.libraries.length; j++)
			{
				var lib = pool.libraries[j];
				var well = lib.bioWell;
				
				var tr = document.createElement('tr');
				if (schema)
				{
					wellIndex = protocol.createMissingRows(tbody, wellsInPool, wellIndex, well, barcodeVariant);
					tr.className = well.column % 2 == 0 ? "evencol" : "oddcol";
				}
				else
				{
					tr.className = 'evencol';
				}
		
				if (barcodeVariant)
				{
					var indexSet = barcodeVariant.indexSets[well.column];
					if (indexSet)
					{
						tr.className = lib && lib.barcode.modified ? 'bg-modified' : indexSet.color;
					}
				}
				var img = '';
				if (lib.isYellow) 
				{
					tr.className += ' yellow-specimen';
					img = yellowImg;
				}
				
				protocol.addColumn(tr, 'lib if-yellow', img+Strings.encodeTags(lib.name));
				protocol.addColumn(tr, 'remain', Numbers.formatNumber((lib.remainingVolume), 1));
				protocol.addColumn(tr, 'molarity', Numbers.formatNumber(lib.molarity, 2));
				if (!libPlate)
				{
					protocol.addColumn(tr, 'workplate', Strings.encodeTags(well.bioPlate.name + ' ' + well.location));
				}
				else if (libPlate.isNeoPrep)
				{
					protocol.addColumn(tr, 'workplate', (8*well.column+well.row+1)+'L');
				}
				else
				{
					protocol.addColumn(tr, 'workplate', Strings.encodeTags(well.location));
				}
				var mixFactor = lib.mixFactor || 1.0;
				protocol.addColumn(tr, 'volume', mixFactor > 1.0 ? '—' : Numbers.formatNumber(lib.actualVolume, 1));
				protocol.addColumn(tr, 'eb', pool.mixingStrategy == 'dynamic' || mixFactor > 1.0 ? '—' : Numbers.formatNumber(lib.actualEb, 1));
				protocol.addColumn(tr, "remarks", lib.remarks.join('; '));
				tbody.appendChild(tr);
			}
			if (schema && wellIndex < wellsInPool.length)
			{
				protocol.createMissingRows(tbody, wellsInPool, wellIndex, null, barcodeVariant);
			}
			
			var poolData = '<div class="pool-data">';
			poolData += pool.libraries.length + ' libs; ';
			poolData += Numbers.formatNumber(pool.molarity, 2)+'nM; ';
			poolData += Numbers.formatNumber(pool.volume, 1) + 'µl';
			poolData += '<br>Mixing strategy: ' + pool.mixingStrategy;
			poolData += '</div>';
			Doc.element('molarity.'+pool.id).innerHTML = poolData;
			Doc.element('eb-volume.'+pool.id).innerHTML = Numbers.formatNumber(Math.max(0, pool.extra.ebVolumeExtra), 1);
			Doc.show('pool.'+pool.id);
		}
		
		Doc.show('all-protocol');
	}
	
	protocol.createMissingRows = function(table, wellsInPool, wellIndex, well, barcodeVariant)
	{
		var wellAtIndex = wellsInPool[wellIndex]; // Empty??
		if (!well) well = {'row': -1, 'column': -1};
		while (wellAtIndex.row != well.row || wellAtIndex.column != well.column)
		{
			var tr = document.createElement('tr');
			tr.className = 'empty';
			if (barcodeVariant)
			{
				var indexSet = barcodeVariant.indexSets[wellAtIndex.column];
				if (indexSet) tr.className += ' ' + indexSet.color;
			}
			
			protocol.addColumn(tr, 'lib', 'not pooled');
			protocol.addColumn(tr, 'remain', '');
			protocol.addColumn(tr, 'molarity', '');
			protocol.addColumn(tr, 'workplate', Reggie.wellToAlpha(wellAtIndex.row) + (wellAtIndex.column+1));
			protocol.addColumn(tr, 'volume', '');
			protocol.addColumn(tr, 'eb', '');
			protocol.addColumn(tr, 'remarks', '');
			table.appendChild(tr);
			wellIndex++;
			if (wellIndex >= wellsInPool.length) break;
			wellAtIndex = wellsInPool[wellIndex];
		}
		return wellIndex+1;
	}
	
	protocol.addColumn = function(tr, className, html)
	{
		var td = document.createElement('td');
		td.className = className;
		td.innerHTML = html;
		tr.appendChild(td);
	}

	
	return protocol;
}();

Doc.onLoad(Protocol.initPage);


var WellPainter = function()
{
	var painter = {};
	
	painter.getClassNameForWell = function(well, schema)
	{
		var cls = '';
		var indexSet = painter.barcodeVariant.indexSets[well.column];
		if (indexSet)
		{
			var lib = well.extract;
			cls += lib && lib.barcode.modified ? 'bg-modified' : indexSet.color;
			if (lib && lib.isYellow) cls += ' yellow-specimen';
		}
		return cls;
	}
	
	painter.getWellText = function(well, schema)
	{
		var text = '';
		var lib = well.extract;
		if (lib)
		{
			var name = lib.name;
			var mixFactor = lib.mixFactor || 1.0;
			var i = name.indexOf('.m');
			var displayName = i == -1 ? Strings.encodeTags(name) : Strings.encodeTags(name.substring(0, i+1))+'<br>&nbsp;'+Strings.encodeTags(name.substring(i));
			text += '<div class="lib if-yellow">'+displayName+'</div>';
			text += '<div><span class="volume">'+Numbers.formatNumber(lib.actualVolume*mixFactor, 1)+'µl</span>';
			if (lib.mixingStrategy == 'fixed' || lib.mixFactor > 1)
			{
				text += '<span class="eb">'+Numbers.formatNumber(lib.actualEb*mixFactor, 1)+'µl</span>';
			}
			else
			{
				text += '<span class="eb">—</span>';
			}
			//text+=lib.eb+':'+lib.mixFactor;
			text += '</div>';
			text += '<div class="remarks">'+ lib.remarks.join('; ') + '</div>';
		}
		return text;
	}

	return painter;
}();


